/* 
RTI-Mage Main Control: Controls lights and camera shutter on RTI-Mage control box to acquire full RTI dataset.
Version 1.1
10/20/2016
Adds mouse commands to Adafruit Bluetooth HID adapter for shutter control with computer

Copyright (C) 2016  Leszek Pawlowicz leszekmp@gmail.com
See the Hackaday.io project page for more info. https://hackaday.io/project/11951-affordable-reflectance-transformation-imaging-dome

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License version 2.0
as published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
 

#include <multiCameraIrControl.h> //Sebastian Setz's library for IR shutter control; http://sebastian.setz.name/arduino/my-libraries/multi-camera-ir-control/
#include <SoftwareSerial.h> //Needed for AdaFruit Bluetooth HID module

//Required for OLED Display
#include <SPI.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>

#define OLED_RESET 4
Adafruit_SSD1306 display(OLED_RESET);

/* This first set of constants should not be changed unless you know exactly what you're doing; system probably won't work 
if any of them are changed. They're used to set the Arduino pin numbers for input/output, and also set essential system constants.*/
const int LED_Pin = A7; //Controls amount of LED on time after shutter is activated. Left knob
const int Photo_Data_Pin=A8; //Controls amount of time LED is off before next picture, for camera to save data. Right knob
const int Beeper_Pin=23; //Direct connection to beeper
const int USB_Camera_Pin=25;  //pin number for voltage to camera USB/IR trigger
const int Servo_Pin=27;  //Controls servo shutter
const int Bluetooth_Tx = 69; //For Bluetooth HID (Pin A15)
const int Bluetooth_Rx = 68; //For Bluetooth HID (Pin A14
const int Action_Pin= 67; //The number of the action button pin (A13)
const int White_Balance_Pin=66; //The number of the white balance pin (A12)
const int Shutter_Switch_Pin=9; //Switch pin to go from USB shutter control to IR shutter control
const int Auto_Switch_Pin=8; // the number for the switch pin to go from manual (off) to auto (on)
const int Beeper_Control_Pin=3; //Connects to switch that enables/disables beeper
const int Source_Pin=38; //First pin to P-Channel MOSFETs (Columns)
const int Sink_Pin=39; //First pin to CAT4101s (Rows)
const int Bounce_Time=250; //min delay time after button is depressed, to work around debounce issues; set this to no less than 250

//Following are constants for the OLED display
String Software_version="1.2";
String Software_date="10/27/2016";
//10/21/16 Added Bluetooth_Shutter subroutine; added left-mouse-click
//10/27/16 Fixed default rows/columns setting
String Current_setting="700mA"; //This must be set here - the system has no way to determine the actual output current, which is set manually.

/*Following constants can be adjusted to set camera times and other parameters. Maximum value for unsigned int constants is 65535 (bit more than a minute);
maximum for unsigned long constants is 4294967295, or roughly 49 days.
All times are in milliseconds, thousandths of a second; 1000 milliseconds = one second, 1500 milliseconds =1.5 milliseconds, etc.*/
const unsigned int Min_Data_Transfer_Time=500;   // minimum time after LED goes off to save photo data
const unsigned int Max_Data_Transfer_Time=6000; // maximum time after LED goes off to save photo data
const unsigned int Min_LED_Time=100; //minimum time LED is on for picture to be taken
const unsigned int Max_LED_Time=4000; // maximum time LED is on for picture to be taken; set by USB microscope time
const unsigned long White_Balance_Time=3000; //time for one white balance LED before switching to the next one
const unsigned long White_Balance_Max_Time=24000; //maximum total time for white balance before it turns off (to prevent LED stress)
const unsigned int LED_Pause_Time=500; // wait time for Bluetooth system to get ready
const unsigned long Min_LED_On_Time=500; //Minimum on time for LED in manual mode
const unsigned long Max_LED_On_Time=4000; //Maximum on time in manual mode for LED (to prevent overheating); turn back on with white balance button. Higher values not recommended.
const unsigned int Frequency=1000; //Sets default frequency of beeper (in Hz, cycles per second)
const unsigned int Rows=8; //Number of Rows being used (cathode connection to CAT4101s)
const unsigned int Columns=8; //Number of Columns being used (anode connection to MOSFET driver outputs)
const int Shutter_Voltage_Time=100; //Time for 5V at USB to either fire CHDK or remote wired shutter.


// Variables will change:


// Variables for shutter time and data transfer time
int LED_Time=Max_LED_Time;
int Data_Transfer_Time=Min_Data_Transfer_Time;
int Current_count; //Counter for LED screen
unsigned long LED_On_Time; // Variable time for LED on in manual mode


/* The following section sets up cameras for use with the IR remote LED. To enable a specific camera, just remove the two slashes "/" in front of the camera make parameters (Canon, Nikon, etc.),
and make sure that there are two slashes in front of every other camera make. You will also need to make a similar change in the IR_Shutter function located further down in the program code */
//Nikon NikonCamera(USB_Camera_Pin);
//String Camera_type="Nikon";
Canon CanonCamera(USB_Camera_Pin);
String Camera_type="Canon";
//CanonWLDC100 CanonWLDC100Camera(USB_Camera_Pin);
//String Camera_type="CanonWLDC100";
//Sony SonyCamera(USB_Camera_Pin);
//String Camera_type="Sony";
//Pentax PentaxCamera(USB_Camera_Pin);
//String Camera_type="Pentax";
//Olympus OlympusCamera(USB_Camera_Pin);
//String Camera_type="Olympus";
//Minolta MinoltaCamera(USB_Camera_Pin);
//String Camera_type="Minolta";

/*Sets digital pin number for Software Serial connection, used with Bluetooth HID module. In this case, digital pin 69 (transmit) is the equivalent of A15 on the Mega2560 board, Dgital pin 68 (A14) is set as receive, even though it's not needed.*/
SoftwareSerial BT = SoftwareSerial(Bluetooth_Rx, Bluetooth_Tx); //Pin A15 is digital 69 Tx, Pin A14 is digital 68 Rx. BT is the output to the Bluetooth module.

//--------------------------------------------------------------------------------------------------

void setup() { //INPUT_PULLUP means high is "off", and eliminates need for pulldown resistor
  pinMode(White_Balance_Pin,INPUT_PULLUP); //Left button; starts white balance
  pinMode(Action_Pin, INPUT_PULLUP); //Right button; starts sequence, also switches rows for white balance
  pinMode(Auto_Switch_Pin, INPUT_PULLUP); //Toggle switch to choose Auto to Manual
  pinMode(Shutter_Switch_Pin,INPUT_PULLUP); //Toggle switch to select CHDK or IR/Bluetooth output
  pinMode(Beeper_Control_Pin,INPUT_PULLUP); //Toggle switch to turn beeper on/off
  pinMode(USB_Camera_Pin,OUTPUT); //Cable output to control CHDK, IR/remote/Bluetooth HID
  pinMode(Beeper_Pin,OUTPUT); //Audio output control
    for (int x=0; x < Columns; x++) {
    pinMode(Source_Pin+x*2,OUTPUT);
    digitalWrite(Source_Pin+x*2,LOW);} //Set P MOSFET pins output mode, low
    for (int y=0; y < Rows; y++) {
    pinMode(Sink_Pin+y*2, OUTPUT);
    digitalWrite(Sink_Pin+y*2,LOW);} //Set CAT4101 current sink pins to output mode, low
  pinMode(13,OUTPUT); //On-board LED
  digitalWrite(13,LOW); // Turn off on-board LED
 // Serial.begin(9600); //Turn on serial communication line
  display.begin(SSD1306_SWITCHCAPVCC, 0x3C);  // initialize OLED display with the I2C addr 0x3C (for the 128x64)
  BT.begin(9600); //Turns on serial output to pin 69, which goes to Bluetooth module.

//Tone sequence when plugged in or reset
if (digitalRead(Beeper_Control_Pin)==HIGH) {
  tone(Beeper_Pin, 1047, 250); 
  delay(200);
  tone(Beeper_Pin, 1369,250);
  delay(200);
  tone(Beeper_Pin, 1568,250);
}
  display.clearDisplay();
  Display_Intro_Text();
}

//---------------------------------------------------------------------------------------------

void loop() {

Display_Standby_Screen();
if (digitalRead(Action_Pin)==LOW && digitalRead(Auto_Switch_Pin)==LOW) {//switch in manual mode
  Manual_mode();}
if (digitalRead(Action_Pin)==LOW && digitalRead(Auto_Switch_Pin)==HIGH) {//switch in auto mode
  Auto_mode();} 
if (digitalRead(White_Balance_Pin)==LOW) {//do white balance sequence / focus test
   delay(Bounce_Time);
   White_Balance();} 
//Display_Text();
}


//---------------------------------------------------------------------------------------------------
void Auto_mode() {  //Loop for running through LEDs automatically
  Current_count=0; // Keeps count of LED # to pass on to OLED display routine

if (digitalRead(Beeper_Control_Pin)==HIGH) {Beep(1,500);} //One long beep to indicate start of automatic mode

  for(int y=0; y < Rows; y++) { //Start of rows loop
      digitalWrite(Sink_Pin+y*2,HIGH); //Turn row on
      for (int x=0; x < Columns; x++) { //Start of loop to cycle through columns
        Current_count++;
        Display_Auto_Screen();
        digitalWrite(Source_Pin+x*2, HIGH); //Turn on LED
        if (digitalRead(Beeper_Control_Pin)==HIGH) {Beep(1,50);}
        if (digitalRead(Shutter_Switch_Pin)==HIGH) { //First part fires IR remote and Bluetooth signal
          BT_Shutter();
          IR_Shutter();} //Fire either IR/Bluetooth shutter ...
          else {USB_Shutter();} //... or USB shutter
        delay(LED_Time); //LED is on for this amount of time
        digitalWrite(Source_Pin+x*2,LOW); //Turns off  LED
        delay(Data_Transfer_Time); //Delay for data transfer in camera
 }//End inner loop

  digitalWrite(Sink_Pin+y*2,LOW); //Turn off row cathode pin before moving on to next row
}
if (digitalRead(Beeper_Control_Pin)==HIGH) Beep(3,200); //Triple beep to announce end of auto
}

//---------------------------------------------------------------------------------------------------


void Manual_mode() {
unsigned long Previous_Millis; //Time markers for shutting off LED if it hasn't been advanced for a while
unsigned long Current_Millis; //Sets start time for monitoring total LED on time

int LED_Status;
  Current_count=0;
  
  for(int y=0; y < Rows; y++) { //Start of loop to cycle through rows
      digitalWrite(Sink_Pin+y*2,HIGH); //Turn row on
      for (int x=0; x < Columns; x++) { //Start of loop to cycle through columns
        Set_Manual_Time();
        Current_count++;
        digitalWrite(Source_Pin+x*2, HIGH); //Turn on bright LED
        LED_Status=1;
        Display_Manual_Screen(LED_Status);
        if (digitalRead(Beeper_Control_Pin)==HIGH) Beep(1,100); //One short beep to signal LED is on
                 
        if (digitalRead(Shutter_Switch_Pin)==HIGH) {
          /*IR/Bluetooth works in manual mode with switch in either USB or IR mode*/
          BT_Shutter();
          IR_Shutter();} //Fire either IR/Bluetooth shutter ...
          else {USB_Shutter();} //... or USB shutter
        Previous_Millis=millis();
        while (digitalRead(Action_Pin)==HIGH) {
          Current_Millis=millis();
          if ((Current_Millis - Previous_Millis) > LED_On_Time) { //Check duration; if over limit, turn light off to prevent burn-out
            digitalWrite(Source_Pin+x*2,LOW);
            Set_Manual_Time();
            LED_Status=0; 
            Display_Manual_Screen(LED_Status);}
          if (digitalRead(White_Balance_Pin)==LOW) {
            digitalWrite(Source_Pin+x*2,HIGH);
            if (digitalRead(Beeper_Control_Pin)==HIGH) {Beep(1,50);}
            Set_Manual_Time();
            LED_Status=1; 
            Display_Manual_Screen(LED_Status);
            delay(Bounce_Time);
            Previous_Millis=Current_Millis;
          }        
        } //Sit and wait until button is pushed again, with light on
        digitalWrite(Source_Pin+x*2,LOW); //Turn off column LED
 }//End inner loop
          digitalWrite(Sink_Pin+y*2,LOW); //Turn off current row cathode pin before moving on to next row
}
if (digitalRead(Beeper_Control_Pin)==HIGH) Beep(3,200); //Three short beeps to indicate end of manual LEDs
  delay(Bounce_Time);

}

//----------------------------------------------------------------------------------------------------

void White_Balance() { //Turn on LEDs at top to set white balance and/or illuminate inside of dome; cycle to reduce wear on one LED

unsigned long Start_Millis=millis(); //Start/End used to turn off white balance after a set period of time, to prevent LED burn & wire overheating
unsigned long End_Millis=millis();
unsigned long Previous_Millis=millis(); //Previous/Current used to time LED on duration
unsigned long Current_Millis=millis();

Beep(1,300); //Sound beeper to indicate start of white balance cycling
delay(Bounce_Time); //Avoid button bounce
int Sink_Pin_Current = Sink_Pin; //Defines which row pin is on
digitalWrite(Sink_Pin,HIGH); //Turn on first row cathode


Start_Loop:
for (int x=0; x < Columns; x++) { //Start of loop to cycle through Columns
End_Millis=millis(); //Current time
if ((End_Millis-Start_Millis) > White_Balance_Max_Time) {goto Exit_Loop;} //Exits white balance after White_Balance_Max_Time exceeded
    digitalWrite(Source_Pin+x*2, HIGH); //Turn on bright LED
    while ((Current_Millis - Previous_Millis) < White_Balance_Time) { //Check to see how much time has passed
    Current_Millis=millis();
      if (digitalRead(White_Balance_Pin)==LOW) {goto Exit_Loop;} //Escape from loop if white balance button is pushed.
      if (digitalRead(Action_Pin)==LOW) { //Switch to next row if action button is pushed
        digitalWrite(Sink_Pin_Current,LOW); //Turn off current row
        delay(Bounce_Time); //Pause for bounce issues
        Sink_Pin_Current=Sink_Pin_Current + 2; //Increment to next row down
        if (Sink_Pin_Current > (Sink_Pin + Rows*2 - 2)) {Sink_Pin_Current=Sink_Pin;} //If already at the bottom row, moves up to top row
        digitalWrite(Sink_Pin_Current,HIGH); //Turn on next row
        Start_Millis=millis(); //Reset time markers after switch to new row
        End_Millis=millis();
        if (digitalRead(Beeper_Control_Pin)==HIGH) Beep(1,100); //Beep to indicate switch to next row
      }
    } 

digitalWrite(Source_Pin+x*2,LOW); //Turn off LED
Previous_Millis=Current_Millis; //Reset 0 time to current time
}
//x=0; //When for loop is done, reset x to 0 to start from beginning
goto Start_Loop;


Exit_Loop:
for (int y=0; y < Rows; y++) { //Start of loop to cycle through Rows
digitalWrite(Sink_Pin+y*2,LOW); //Make sure all row voltages are off
}
for (int x=0; x < Columns; x++) { //Start of loop to cycle through Columns
digitalWrite(Source_Pin+x*2,LOW); //Make sure all column voltages are off
}
if (digitalRead(Beeper_Control_Pin)==HIGH) Beep(3,200); //Beep to indicate white balance turned off
}

//------------------------------------------------------------------------------------------------------

void Set_Data_Transfer_Time() { //Sets data transfer time delay from potentiometer
int Data_Transfer_Sensor;

Data_Transfer_Sensor=  analogRead(Photo_Data_Pin);
Data_Transfer_Time= map(Data_Transfer_Sensor,0,1023,Min_Data_Transfer_Time,Max_Data_Transfer_Time);
}

//-----------------------------------------------------------------------------------------------------
void Set_LED_Time() { //Sets LED on time from potentiometer
int LED_Sensor;

LED_Sensor=  analogRead(LED_Pin);
LED_Time= map(LED_Sensor,0,1023,Min_LED_Time,Max_LED_Time);
}
//------------------------------------------------------------------------------------------------------
void Set_Manual_Time() { //Sets manual LED time from potentiometer
int LED_Sensor;

LED_Sensor=  analogRead(LED_Pin);
LED_On_Time= map(LED_Sensor,0,1023,Min_LED_On_Time,Max_LED_On_Time);//Limits for LED on time in manual mode
}
//------------------------------------------------------------------------------------------------------

void USB_Shutter() { //Controls shutter for CHDK-enabled cameras directly through USB cable connection
Set_Data_Transfer_Time();
Set_LED_Time();

digitalWrite(USB_Camera_Pin,HIGH); //Turn on 5V to USB Shutter cable
delay(Shutter_Voltage_Time); //5V Time to fire camera
digitalWrite(USB_Camera_Pin,LOW); //Turns off 5V to USB Shutter cable

}

//------------------------------------------------------------------------------------------------------

void IR_Shutter() { //Controls shutter for IR-controlled cameras through IR LED attached to USB cable
Set_Data_Transfer_Time();
Set_LED_Time();
        
//Section for IR remotes        
//Remove double-slash ("//") from front of camera make being used to activate; place double-slash in front of unneeded cameras to de-activate. Make sure you've also done this in the initialization section at top.
//NikonCamera.shutterNow();
CanonCamera.shutterNow();
//CanonWLDC100Camera.shutterNow();
//SonyCamera.shutterNow();
//PentaxCamera.shutterNow();
//OlympusCamera.shutterNow();;
//MinoltaCamera.shutterNow();

}
//________________________________________________________________________________________________
void BT_Shutter() { //Sets commands for Adafruit Bluetooth HID module to fire shutter using computer
          //Fires shutter with Alt-f, s command
          //BT.write(0xE2); //Alt key
          //BT.write(0x66); //f key
          //delay(25);
          //BT.write(0x73); //s key

          //Fires shutter with left mouse button
          delay(LED_Pause_Time);
          mouseCommand(0x1,0,0); //Left button down
          delay(25);
          mouseCommand(0,0,0); //Release button
}

//------------------------------------------------------------------------------------------------------
void mouseCommand(uint8_t buttons, uint8_t x, uint8_t y) { //Sends mouse commands from Adafruit HID Bluetooth module
  BT.write(0xFD);
  BT.write((byte)0x00);
  BT.write((byte)0x03);
  BT.write(buttons);
  BT.write(x);
  BT.write(y);
  BT.write((byte)0x00);
  BT.write((byte)0x00);
  BT.write((byte)0x00);
  
}
 //--------------------------------------------------------------------------------------------------------
 
void Beep(int Beep_Repeat,int Beep_Length ) { //Beep subroutine; Beep_Repeat = # of beeps, Beep_Length = length of beep and gap between beeps

tone(Beeper_Pin, Frequency, Beep_Length); //First beep

for (int i=0; i < (Beep_Repeat-1); i++) {  
  delay(Beep_Length); //Gap between beeps
  tone(Beeper_Pin, Frequency, Beep_Length); //Turns on beeper for desired length
}
}

//---------------------------------------------------------------------------------------------------------------------------------------
void Display_Intro_Text() {  //Shows intro OLED screen

  String Col_string=String(Columns);
  String Row_string=String(Rows);
  String LED_total=String(Columns*Rows);

  display.setTextSize(3);
  display.setTextColor(WHITE);
  display.setCursor(0,0);
  display.println("RTIMage");
  display.setTextSize(1);
  display.println("");
  display.println("Ver. " + Software_version + "  " + Software_date);
  display.println("by Leszek Pawlowicz");
  display.println("Row: " + Row_string + " Col: " + Col_string + " LED: " + LED_total);
  display.println("LED Current: " + Current_setting);
  display.display();
  delay(5000);
  display.clearDisplay();
}
//---------------------------------------------------------------------------------------------------------------------------------

void Display_Standby_Screen() {  //Shows standby OLED screen


  Set_Data_Transfer_Time();
  Set_LED_Time();
  Set_Manual_Time();
  String Shutter_string=String((float) LED_Time/1000, 2);
  String Data_Transfer_string=String((float) Data_Transfer_Time/1000, 2);
  String LED_On_Time_string = String((float) LED_On_Time/1000,2);
  
  display.clearDisplay();
  display.setTextSize(1);
  display.setTextColor(WHITE);
  display.setCursor(0,0);

  if (digitalRead(Beeper_Control_Pin)==HIGH) {
    display.println("Sound: On"); 
  }
  else {display.println("Sound: Off");}
  if (digitalRead(Auto_Switch_Pin)==LOW) {
    display.println("Mode: Manual");
    display.println("");
    display.println("");}
  if (digitalRead(Auto_Switch_Pin)==HIGH) {
    display.println("Mode: Auto");
    if (digitalRead(Shutter_Switch_Pin)==HIGH) {
      display.println("Shutter: IR/Bluetooth");
      display.println("Camera: " + Camera_type);
    }
    else {display.println("Shutter: USB");
    display.println("");}
    } 
    display.setTextSize(2);
  if (digitalRead(Auto_Switch_Pin)==LOW) {
   display.println("LED: " + LED_On_Time_string + "s");
   display.println("Del: NA"); 
  }
  else {display.println("LED: " + Shutter_string + "s");
  display.println("Del: " + Data_Transfer_string + "s");}
  display.setTextSize(1);
  display.display();
  display.clearDisplay();
}
//-----------------------------------------------------------------------------------------------------------------------------------------------------
void Display_Auto_Screen() {  //Shows OLED screen in Automatic Mode

  String Current_count_string=String(Current_count);
  String Total_LEDs_string=String(Rows*Columns);
  Set_Data_Transfer_Time();
  Set_LED_Time();

  
  String Shutter_string=String((float) LED_Time/1000, 2);
  String Data_Transfer_string=String((float) Data_Transfer_Time/1000, 2);
 
  display.clearDisplay();
  display.setTextSize(1);
  display.setTextColor(WHITE);
  display.setCursor(0,0); 
  display.setTextSize(3);
  if (Current_count < 10) {
    display.println("  " + Current_count_string + "/" + Total_LEDs_string);
  }
  else {display.println(" " +Current_count_string + "/" + Total_LEDs_string);}
  display.setTextSize(1);
  display.println("");
  display.setTextSize(2);
  display.println("LED: " + Shutter_string + "s");
  display.println("Del: " + Data_Transfer_string + "s");
  display.display();
  display.clearDisplay();
}

//---------------------------------------------------------------------------------------------------------------
void Display_Manual_Screen(int LED_Status) {  //Shows OLED screen in manual mode

  String Current_count_string=String(Current_count);
  String Total_LEDs_string=String(Rows*Columns);
  String LED_On_Time_string = String((float) LED_On_Time/1000,2);
  
  display.clearDisplay();
  display.setTextSize(1);
  display.setTextColor(WHITE);
  display.setCursor(0,0);

  
  display.setTextSize(3);
  if (Current_count < 10) {
    display.println("  " + Current_count_string + "/" + Total_LEDs_string);
  }
  else {display.println(" " +Current_count_string + "/" + Total_LEDs_string);}
  display.setTextSize(1);
  display.println("");
  display.setTextSize(2);
  if (LED_Status ==1) {display.println("LED: On");}
    else {display.println("LED: Off");}
  display.println("Time:" + LED_On_Time_string + "s");
  display.setTextSize(1);
  display.display();
  display.clearDisplay();
}
